@@ -8,6 +8,7 @@ import sys
 from .utils import flatten
 from .errorlog import ErrorLog, errorlog
 from .imagemanager import image_manager
+from .logmanager import log_manager
 
 
 class DocsGenException(Exception):
@@ -28,6 +29,7 @@ class GenericBlock(object):
         self.parent = parent
         self.children = []
         self.figure_num = 0
+        self.definitions = {}
         if parent:
             parent.children.append(self)
 
@@ -82,24 +84,31 @@ class GenericBlock(object):
             d["children"] = [child.get_data() for child in self.children]
         return d
 
-    def get_link(self, target, currfile=None):
+    def get_link(self, target, currfile=None, literalize=False, html=False):
         return self.title
 
-    def parse_links(self, line, controller, target):
+    def parse_links(self, line, controller, target, html=False):
         oline = ""
         while line:
             m = self._link_pat.match(line)
             if m:
                 oline += m.group(1)
-                name = m.group(2)
+                name = m.group(2).lower().strip()
                 line = m.group(3)
-                if name not in controller.items_by_name:
+                literalize = name.endswith("()")
+                if name in controller.items_by_name:
+                    item = controller.items_by_name[name]
+                    oline += item.get_link(target, currfile=self.origin.file, literalize=literalize, html=html)
+                elif name in controller.definitions:
+                    oline += target.get_link(name, anchor=name.lower(), file="Glossary", literalize=literalize, html=html)
+                elif name in controller.defn_aliases:
+                    term = controller.defn_aliases[name]
+                    oline += target.get_link(name, anchor=term.lower(), file="Glossary", literalize=literalize, html=html)
+                else:
+                    print(controller.definitions)
                     msg = "Invalid Link {{{{{0}}}}}".format(name)
                     errorlog.add_entry(self.origin.file, self.origin.line, msg, ErrorLog.FAIL)
                     oline += name
-                else:
-                    item = controller.items_by_name[name]
-                    oline += item.get_link(target, currfile=self.origin.file)
             else:
                 oline += line
                 line = ""
@@ -241,6 +250,37 @@ class TextBlock(GenericBlock):
         super().__init__(title, subtitle, body, origin, parent=parent)
 
 
+class DefinitionsBlock(GenericBlock):
+    def __init__(self, title, subtitle, body, origin, parent=None):
+        super().__init__(title, subtitle, body, origin, parent=parent)
+        terms = []
+        self.definitions = {}
+        for line in body:
+            if '=' not in line:
+                raise DocsGenException(title, "Expected body line in the format TERM = DEFINITION, while declaring block:")
+            termset, defn = line.split('=', 1)
+            termset = [x.strip() for x in termset.lower().split('|')]
+            defn = defn.strip()
+            for term in termset:
+                if term in terms:
+                    raise DocsGenException(title, "Redefined term '{}', while declaring block:".format(term))
+                terms.append(term)
+            term = termset[0]
+            self.definitions[term] = (termset, defn)
+
+    def get_file_lines(self, controller, target):
+        out = target.block_header(self.title, self.subtitle)
+        terms = list(self.definitions.keys())
+        defs = {
+                key: self.parse_links(info[1], controller, target, html=True)
+                for key, info in self.definitions.items()
+            }
+        for term in terms:
+            out.extend(target.markdown_block(["{}: {}".format(term.title(), defs[term])]))
+        out.append("")
+        return out
+
+
 class BulletListBlock(GenericBlock):
     def __init__(self, title, subtitle, body, origin, parent=None):
         super().__init__(title, subtitle, body, origin, parent=parent)
@@ -330,7 +370,7 @@ class FileBlock(GenericBlock):
         d["children"] = list(filter(lambda x: x["name"] not in skip_titles, d["children"]))
         return d
 
-    def get_link(self, target, currfile=None, label=""):
+    def get_link(self, target, currfile=None, label="", literalize=False, html=False):
         file = self.origin.file
         if currfile is None or self.origin.file == currfile:
             file = ""
@@ -338,7 +378,8 @@ class FileBlock(GenericBlock):
             label=label if label else str(self),
             anchor="",
             file=file,
-            literalize=False
+            literalize=literalize,
+            html=html
         )
 
     def get_tocfile_lines(self, controller, target, n=1, currfile=""):
@@ -421,7 +462,7 @@ class SectionBlock(GenericBlock):
         if parent:
             self.parent.figure_num += 1
 
-    def get_link(self, target, currfile=None, label=""):
+    def get_link(self, target, currfile=None, label="", literalize=False, html=False):
         file = self.origin.file
         if currfile is None or self.origin.file == currfile:
             file = ""
@@ -429,7 +470,8 @@ class SectionBlock(GenericBlock):
             label=label if label else str(self),
             anchor=target.header_link(str(self)),
             file=file,
-            literalize=False
+            literalize=literalize,
+            html=html
         )
 
     def get_tocfile_lines(self, controller, target, n=1, currfile=""):
@@ -551,7 +593,7 @@ class SubsectionBlock(GenericBlock):
         if parent:
             self.parent.figure_num += 1
 
-    def get_link(self, target, currfile=None, label=""):
+    def get_link(self, target, currfile=None, label="", literalize=False, html=False):
         file = self.origin.file
         if currfile is None or self.origin.file == currfile:
             file = ""
@@ -559,7 +601,8 @@ class SubsectionBlock(GenericBlock):
             label=label if label else str(self),
             anchor=target.header_link(str(self)),
             file=file,
-            literalize=False
+            literalize=literalize,
+            html=html
         )
 
     def get_tocfile_lines(self, controller, target, n=1, currfile=""):
@@ -667,7 +710,7 @@ class ItemBlock(LabelBlock):
             re.sub(r'\([^\)]*\)', r'()', self.subtitle)
         )
 
-    def get_link(self, target, currfile=None, label="", literalize=True):
+    def get_link(self, target, currfile=None, label="", literalize=True, html=False):
         file = self.origin.file
         if currfile is None or self.origin.file == currfile:
             file = ""
@@ -677,7 +720,8 @@ class ItemBlock(LabelBlock):
                 "{}: {}".format(self.title, self.subtitle)
             ),
             file=file,
-            literalize=literalize
+            literalize=literalize,
+            html=html
         )
 
     def get_data(self):
@@ -819,8 +863,66 @@ class ItemBlock(LabelBlock):
         return "{}{}".format(hdr, self.figure_num)
 
 
+class LogBlock(GenericBlock):
+    def __init__(self, title, subtitle, body, origin, parent=None, meta=""):
+        super().__init__(title, subtitle, body, origin, parent=parent)
+        self.meta = meta
+        self.log_output = []
+        self.log_title = subtitle if subtitle.strip() else "Log Output"
+
+        fileblock = parent
+        while fileblock.parent:
+            fileblock = fileblock.parent
+
+        script_lines = []
+        script_lines.extend(fileblock.includes)
+        script_lines.extend(fileblock.common_code)
+        for line in self.body:
+            if line.strip().startswith("--"):
+                script_lines.append(line.strip()[2:])
+            else:
+                script_lines.append(line)
+        self.raw_script = script_lines
+        self.generate_log()
+
+    def generate_log(self):
+        self.log_request = log_manager.new_request(
+            self.origin.file, self.origin.line,
+            self.raw_script,
+            starting_cb=self._log_proc_start,
+            completion_cb=self._log_proc_done,
+            verbose=True  
+        )
+        log_manager.process_requests()
+
+    def _log_proc_start(self, req):
+        print("  Processing log for {}:{}... ".format(self.origin.file, self.origin.line), end='')
+        sys.stdout.flush()
+
+    def _log_proc_done(self, req):
+        if req.success:
+            self.log_output = req.echos
+            print("SUCCESS")
+        else:
+            self.log_output = []
+            #print("FAIL")
+            print("FAIL: " + "\n".join(req.errors + req.warnings))
+        sys.stdout.flush()
+
+    def get_file_lines(self, controller, target):
+        out = []
+        #if self.log_output:
+        if self.log_request.success and self.log_request.echos:    
+            #out.extend(target.block_header("Log Output", ""))
+            out.extend(target.block_header(self.log_title, ""))
+            # out.extend(target.markdown_block(["```log"] + self.log_output + ["```"]))
+            out.extend(target.markdown_block(["```log"] + self.log_request.echos + ["```"]))
+        else:
+            print(f"WARNING: No log output for {self.origin.file}:{self.origin.line}")    
+        return out
+
 class ImageBlock(GenericBlock):
-    def __init__(self, title, subtitle, body, origin, parent=None, meta="", use_apngs=False):
+    def __init__(self, title, subtitle, body, origin, verbose=False, enabled_features=[], parent=None, meta="", use_apngs=False):
         super().__init__(title, subtitle, body, origin, parent=parent)
         fileblock = parent
         while fileblock.parent:
@@ -865,8 +967,10 @@ class ImageBlock(GenericBlock):
         file_base = os.path.splitext(file_name)[0]
         self.image_url_rel = os.path.join("images", file_base, proposed_name)
         self.image_url = os.path.join(file_dir, self.image_url_rel)
+        self.verbose = verbose
+        self.enabled_features = enabled_features
 
-    def generate_image(self, target):
+    def generate_image(self, target, parser=None):
         self.image_req = None
         if "NORENDER" in self.meta:
             return
@@ -879,12 +983,15 @@ class ImageBlock(GenericBlock):
             outfile = os.path.join(target.docs_dir, self.image_url)
             outdir = os.path.dirname(outfile)
             os.makedirs(outdir, mode=0o744, exist_ok=True)
-
+            default_colorscheme = parser.default_colorscheme if parser else "Cornfield"
             self.image_req = image_manager.new_request(
                 self.origin.file, self.origin.line,
                 outfile, self.raw_script, self.meta,
                 starting_cb=self._img_proc_start,
-                completion_cb=self._img_proc_done
+                completion_cb=self._img_proc_done,
+                verbose=self.verbose,
+                enabled_features=self.enabled_features,
+                default_colorscheme=default_colorscheme
             )
 
     def get_data(self):
@@ -930,7 +1037,7 @@ class ImageBlock(GenericBlock):
         if "Hide" in self.meta:
             return out
 
-        self.generate_image(target)
+        self.generate_image(target, controller)
 
         code = []
         code.extend([line for line in fileblock.includes])
@@ -960,13 +1067,13 @@ class ImageBlock(GenericBlock):
 
 
 class FigureBlock(ImageBlock):
-    def __init__(self, title, subtitle, body, origin, parent, meta="", use_apngs=False):
-        super().__init__(title, subtitle, body, origin, parent=parent, meta=meta, use_apngs=use_apngs)
+    def __init__(self, title, subtitle, body, origin, parent, verbose=False, enabled_features=[], meta="", use_apngs=False):
+        super().__init__(title, subtitle, body, origin, verbose=verbose, enabled_features=enabled_features, parent=parent, meta=meta, use_apngs=use_apngs)
 
 
 class ExampleBlock(ImageBlock):
-    def __init__(self, title, subtitle, body, origin, parent, meta="", use_apngs=False):
-        super().__init__(title, subtitle, body, origin, parent=parent, meta=meta, use_apngs=use_apngs)
+    def __init__(self, title, subtitle, body, origin, parent, verbose=False, enabled_features=[], meta="", use_apngs=False):
+        super().__init__(title, subtitle, body, origin, verbose=verbose, enabled_features=enabled_features, parent=parent, meta=meta, use_apngs=use_apngs)
 
 
 
