@@ -9,6 +9,7 @@ import glob
 from .errorlog import ErrorLog, errorlog
 from .imagemanager import image_manager
 from .blocks import *
+from .logmanager import log_manager
 from .filehashes import FileHashes
 
 
@@ -44,12 +45,16 @@ class DocsGenParser(object):
         self.priority_files = []
         self.priority_groups = []
         self.items_by_name = {}
+        self.definitions = {}
+        self.defn_aliases = {}
         self.syntags_data = {}
+        self.default_colorscheme = "Cornfield"
 
         sfx = self.target.get_suffix()
         self.TOCFILE = "TOC" + sfx
         self.TOPICFILE = "Topics" + sfx
         self.INDEXFILE = "AlphaIndex" + sfx
+        self.GLOSSARYFILE = "Glossary" + sfx
         self.CHEATFILE = "CheatSheet" + sfx
         self.SIDEBARFILE = "_Sidebar" + sfx
 
@@ -85,7 +90,18 @@ class DocsGenParser(object):
         aliases = [x.strip() for x in subtitle.split(",")]
         self.curr_item.aliases.extend(aliases)
         for alias in aliases:
-            self.items_by_name[alias] = self.curr_item
+            self.items_by_name[alias.lower()] = self.curr_item
+
+    def _validate_colorscheme(self, colorscheme):
+        """Validate the color scheme against OpenSCAD's supported schemes."""
+        valid_schemes = [
+            'Cornfield', 'Metallic', 'Sunset', 'Starnight', 'ClearSky', 'BeforeDawn', 'Nature', 'Daylight Gem', 'Nocturnal Gem', 
+            'DeepOcean', 'Solarized', 'Tomorrow', 'Tomorrow Night' 
+        ]
+        if colorscheme not in valid_schemes:
+            errorlog.add_entry(self.RCFILE, 0, f"Invalid ColorScheme '{colorscheme}' in {self.RCFILE}. Using 'Cornfield'.", ErrorLog.WARNING)
+            return 'Cornfield'
+        return colorscheme
 
     def _skip_lines(self, lines, line_num=0):
         while line_num < len(lines):
@@ -200,6 +216,14 @@ class DocsGenParser(object):
             origin = OriginInfo(src_file, hdr_line_num+1)
             if title == "DefineHeader":
                 self._define_blocktype(subtitle, meta)
+            elif title == "ColorScheme":
+                if origin.file != self.RCFILE:
+                    raise DocsGenException(title, f"Block disallowed outside of {self.RCFILE} file:")
+                if body:
+                    raise DocsGenException(title, "Body not supported, while declaring block:")
+                if not subtitle:
+                    raise DocsGenException(title, "Must provide a color scheme (e.g., Tomorrow), while declaring block:")
+                self.default_colorscheme = self._validate_colorscheme(subtitle.strip())
             elif title == "IgnoreFiles":
                 if origin.file != self.RCFILE:
                     raise DocsGenException(title, "Block disallowed outside of {} file:".format(self.RCFILE))
@@ -226,6 +250,9 @@ class DocsGenParser(object):
                     raise DocsGenException(title, "Body not supported, while declaring block:")
                 self.opts.docs_dir = subtitle.strip().rstrip("/")
                 self.opts.update_target()
+            elif title == "EnabledFeatures":
+                self.opts.enabled_features = [item.strip() for item in subtitle.split(",") if item.strip()]
+                self.opts.update_target()
             elif title == "UsePNGAnimations":
                 if origin.file != self.RCFILE:
                     raise DocsGenException(title, "Block disallowed outside of {} file:".format(self.RCFILE))
@@ -255,8 +282,13 @@ class DocsGenParser(object):
                 if body:
                     raise DocsGenException(title, "Body not supported, while declaring block:")
                 if not (
-                    self.opts.gen_files or self.opts.gen_toc or self.opts.gen_index or
-                    self.opts.gen_topics or self.opts.gen_cheat or self.opts.gen_sidebar
+                    self.opts.gen_files or
+                    self.opts.gen_toc or
+                    self.opts.gen_index or
+                    self.opts.gen_topics or
+                    self.opts.gen_cheat or
+                    self.opts.gen_sidebar or
+                    self.opts.gen_glossary
                 ):
                     # Only use default GeneratedDocs if the command-line doesn't specify any docs
                     # types to generate.
@@ -273,6 +305,8 @@ class DocsGenParser(object):
                             self.opts.gen_topics = True
                         elif part in ["CHEAT", "CHEATSHEET"]:
                             self.opts.gen_cheat = True
+                        elif part == "GLOSSARY":
+                            self.opts.gen_glossary = True
                         elif part == "SIDEBAR":
                             self.opts.gen_sidebar = True
                         else:
@@ -361,22 +395,36 @@ class DocsGenParser(object):
             elif title == "CommonCode":
                 self._check_filenode(title, origin)
                 self.curr_file_block.common_code.extend(body)
+            elif title == "Definitions":
+                self._check_filenode(title, origin)
+                block = DefinitionsBlock(title, subtitle, body, origin, parent=parent)
+                for main_term, info in block.definitions.items():
+                    terms, defn = info
+                    for term in terms:
+                        if term.lower() in self.definitions or term in self.defn_aliases:
+                            raise DocsGenException(title, 'Term "{}" re-defined, while declaring block:'.format(term))
+                    self.definitions[main_term.lower()] = (terms, defn)
+                    for term in terms[1:]:
+                        self.defn_aliases[term.lower()] = main_term
             elif title == "Figure":
                 self._check_filenode(title, origin)
-                FigureBlock(title, subtitle, body, origin, parent=parent, meta=meta, use_apngs=self.opts.png_animation)
+                FigureBlock(title, subtitle, body, origin, verbose=self.opts.verbose, parent=parent, enabled_features=self.opts.enabled_features, meta=meta, use_apngs=self.opts.png_animation)
             elif title == "Example":
                 if self.curr_item:
-                    ExampleBlock(title, subtitle, body, origin, parent=parent, meta=meta, use_apngs=self.opts.png_animation)
+                    ExampleBlock(title, subtitle, body, origin, verbose=self.opts.verbose, parent=parent, enabled_features=self.opts.enabled_features, meta=meta, use_apngs=self.opts.png_animation)
             elif title == "Figures":
                 self._check_filenode(title, origin)
                 for lnum, line in enumerate(body):
-                    FigureBlock("Figure", subtitle, [line], origin, parent=parent, meta=meta, use_apngs=self.opts.png_animation)
+                    FigureBlock("Figure", subtitle, [line], origin, verbose=self.opts.verbose, parent=parent, enabled_features=self.opts.enabled_features, meta=meta, use_apngs=self.opts.png_animation)
                     subtitle = ""
             elif title == "Examples":
                 if self.curr_item:
                     for lnum, line in enumerate(body):
-                        ExampleBlock("Example", subtitle, [line], origin, parent=parent, meta=meta, use_apngs=self.opts.png_animation)
+                        ExampleBlock("Example", subtitle, [line], origin, verbose=self.opts.verbose, enabled_features=self.opts.enabled_features, parent=parent, meta=meta, use_apngs=self.opts.png_animation)
                         subtitle = ""
+            elif title == "Log":
+                if self.curr_item:
+                    LogBlock(title, subtitle.strip(), body, origin, parent=parent, meta=meta)                     
             elif title in self.header_defs:
                 parcls, cls, data, cb = self.header_defs[title]
                 if not parcls or isinstance(self.curr_parent, parcls):
@@ -399,7 +447,7 @@ class DocsGenParser(object):
                     msg = "Previous declaration of `{}` at {}:{}, Redeclared:".format(subtitle, prevorig.file, prevorig.line)
                     raise DocsGenException(title, msg)
                 item = ItemBlock(title, subtitle, body, origin, parent=parent)
-                self.items_by_name[subtitle] = item
+                self.items_by_name[subtitle.lower()] = item
                 self.curr_item = item
                 self.curr_parent = item
             elif title == "Synopsis":
@@ -672,6 +720,11 @@ class DocsGenParser(object):
                 col = 1
             self.parse_file(filename, commentless=commentless)
             col = col + flen
+        for key, info in self.definitions.items():
+            keys, defn = info
+            blk = self.file_blocks[0]
+            defn = blk.parse_links(defn, self, self.target, html=False)
+            self.definitions[key.lower()] = (keys, defn)
         if not self.quiet:
             print("")
 
@@ -783,6 +836,47 @@ class DocsGenParser(object):
             for line in out:
                 f.write(line + "\n")
 
+    def write_glossary_file(self):
+        """Generates the Glossary file from the parsed documentation."""
+        target = self.opts.target
+        os.makedirs(target.docs_dir, mode=0o744, exist_ok=True)
+        defs = {key: info[1] for key, info in self.definitions.items()}
+        sorted_words = sorted(list(defs.keys()), key=lambda v: v.upper())
+        ltrs_found = {}
+        for word in sorted_words:
+            ltr = word[0].upper()
+            ltrs_found[ltr] = 1
+        ltrs_found = sorted(list(ltrs_found.keys()))
+
+        out = []
+        out = target.header("Glossary")
+        out.extend(target.markdown_block([
+            "Definitions of various Words and terms."
+        ]))
+        out.extend(target.markdown_block([
+            "  ".join(
+                target.get_link(ltr.upper(), anchor=ltr.lower(), literalize=False)
+                for ltr in ltrs_found
+            )
+        ]))
+        out.extend(target.horizontal_rule())
+        old_ltr = ''
+        for word in sorted_words:
+            ltr = word[0].upper()
+            if old_ltr != ltr:
+                out.extend(target.header(ltr.upper(), lev=2))
+                old_ltr = ltr
+            defn = defs[word.lower()]
+            out.extend(target.header(word.title(), lev=3))
+            out.extend(target.markdown_block([defn]))
+        out = target.postprocess(out)
+        outfile = os.path.join(target.docs_dir, self.GLOSSARYFILE)
+        if not self.quiet:
+            print("Writing {}...".format(outfile))
+        with open(outfile, "w") as f:
+            for line in out:
+                f.write(line + "\n")
+
     def write_topics_file(self):
         """Generates the Topics file from the parsed documentation."""
         target = self.opts.target
@@ -944,6 +1038,8 @@ class DocsGenParser(object):
             out.extend(target.line_with_break(target.get_link("Function Index", file="AlphaIndex", literalize=False)))
         if self.opts.gen_topics:
             out.extend(target.line_with_break(target.get_link("Topics Index", file="Topics", literalize=False)))
+        if self.opts.gen_glossary:
+            out.extend(target.line_with_break(target.get_link("Glossary", file="Glossary", literalize=False)))
         if self.opts.gen_cheat:
             out.extend(target.line_with_break(target.get_link("Cheat Sheet", file="CheatSheet", literalize=False)))
         if self.opts.sidebar_middle:
